package edu.berkeley.sbp;
import edu.berkeley.sbp.*;
import edu.berkeley.sbp.util.*;
import edu.berkeley.sbp.bind.*;
import java.io.*;
import java.util.*;
import java.lang.reflect.*;

/** a tree (or node in a tree); see jargon.txt for details */
public class Tree<T>
    extends PrintableTree<Tree<T>>
    implements Iterable<Tree<T>>,
               GraphViz.ToGraphViz {

    private final Input.Region location;
    private final T            head;
    private final Tree<T>[]    children;
    private final boolean      lift;

    public T                 head()        { return head; }

    private Tree<T> lifted() { return children[children.length-1]; }

    public int               numChildren() {
        return lift
            ? (children.length-1)+lifted().numChildren()
            : children.length;
    }
    public Iterable<Tree<T>> children()    { return this; }
    public Iterator<Tree<T>> iterator()    {
        return lift
            ? new ConcatenateIterator(new ArrayIterator(children, 0, children.length-1),
                                      children[children.length-1].iterator())
            : new ArrayIterator(children);
    }

    public Tree<T> child(int i)  {
        return lift && i >= children.length-1
            ? children[children.length-1].child(i-(children.length-1))
            : children[i];
    }

    public Input.Region    getRegion() { return location; }

    public Tree(Input.Region loc, T head)                     { this(loc, head, null); }
    public Tree(Input.Region loc, T head, Tree<T>[] children) { this(loc, head, children, false); }
    public Tree(Input.Region loc, T head, Tree<T>[] children, boolean lift) {
        this.location = loc;
        this.head = head;
        this.lift = lift && children != null && children.length > 0;
        this.children = ArrayUtil.clone(children, Tree.class);
    }


    // PrintableTree /////////////////////////////////////////////////////////////////////////////

    protected String headToString() { return head()==null?null:head().toString(); }
    protected String headToJava()   {
      // FIXME
        if (head()==null) return null;
        if (head() instanceof ToJava) {
            StringBuffer sb = new StringBuffer();
            ((ToJava)head()).toJava(sb);
            return sb.toString();
        }
        return (head()==null?"null":("\""+StringUtil.toJavaString(head().toString())+"\""));
    }
    protected String left()   { return "{"; }
    protected String right()  { return "}"; }
    protected boolean ignoreSingleton() { return false; }


    // ToGraphViz /////////////////////////////////////////////////////////////////////////////

    public GraphViz.Node toGraphViz(GraphViz gv) {
        if (gv.hasNode(this)) return gv.createNode(this);
        GraphViz.Node n = gv.createNode(this);
        n.label = head()==null ? "" : head().toString();
        //n.color = "red";
        for(Tree t : this) n.edge(t, null);
        return n;
    }
    public boolean isTransparent() { return false; }
    public boolean isHidden() { return false; }


    // TreeFunctor /////////////////////////////////////////////////////////////////////////////

    public static interface TreeFunctor<T,R> extends Functor<Iterable<Tree<T>>, R> { }
    
}
